from os import environ, path, access, W_OK
from pathlib import Path
from io import StringIO
from boto3.session import Session as botosession
from botocore.configloader import load_config
from botocore.session import get_session
from .providers import CognitoIdentity, TokenFetcher, TokenCache

COGNITO_DEFAULT_SESSION = None


def _get_default_session(**kwargs):
    global COGNITO_DEFAULT_SESSION
    if COGNITO_DEFAULT_SESSION is None:
        COGNITO_DEFAULT_SESSION = Session(**kwargs)
    return COGNITO_DEFAULT_SESSION


def Session(**kwargs):

    args_to_remove = [
        "cognito_profile",
        "cognito_config",
        "cognito_credentials_file",
        "auth_type",
        "token_cache"
    ]

    # Resolve profile through possible config file locations and then config
    if kwargs.get("cognito_profile") and kwargs.get("cognito_config"): raise ValueError("kwargs cognito_profile and cognito_config are mutually exclusive")

    if profile_name := kwargs.get("cognito_profile"):
        config = get_profile(profile_name, kwargs.get("cognito_credentials_file"))
    elif kwargs.get("cognito_config"):
        config = kwargs["cognito_config"]
    else:
        config = {}

    auth_type = kwargs.get("auth_type") or config.get("auth_type") or "user_srp"
    if auth_type not in ("user_passwword", "user_srp"): raise ValueError("auth_type must be one of user_password or user_srp")

    # If token_cache file is provided we will use it, otherwise we will use a StringIO object
    if token_cache := kwargs.get("token_cache"):
        if not path.isfile(token_cache): raise FileNotFoundError(f"File {token_cache} specified as a token cache does not exist.")
        if not access(token_cache, W_OK): raise OSError(f"File {token_cache} is not writable.")
        token_cache = TokenCache(token_cache)
    else:
        cache_io = StringIO()
        token_cache = TokenCache(cache_io)

    # Create our credential provider
    auth_client = CognitoIdentity(
        auth_type,
        config=config,
        token_cache=token_cache
    )

    bc_session = get_session()

    # Clean our own kwargs out so that we can pass the remainder to boto.
    session_args = {k: kwargs[k] for k in kwargs if k not in args_to_remove}
    session_args["botocore_session"] = bc_session

    session = botosession(**session_args)

    # Now we set the session so when we get the token's getter properties we
    # retreive from the same cache that the provider is using.
    session.token_cache = token_cache

    # Now we can fire a call to refresh auth tokens if we need to
    session.refresh_auth = auth_client.refresh_auth

    # Insert our credential provider at the top of the chain
    cognito_provider = bc_session.get_component('credential_provider')
    cognito_provider.insert_before('env', auth_client)

    # If we don't pre-load then we won't have access to the tokens until after either client() or resource() is called
    auth_client.load()
    return session


def get_profile(name, credentials_file):
    credentials_file = credentials_file or environ.get("COGNITO_CREDENTIALS_FILE") or f"{Path.home()}/.aws/cognito_credentials"
    if not path.isfile(credentials_file): raise Exception("Cannot find cognito_credentials_file")
    profiles = parse_config(credentials_file).get("profiles", {})
    if name not in profiles:
        raise Exception(f"Specified cognito_profile does not exist in {credentials_file}")
    return profiles[name]


def client(*args, **kwargs):
    session_opts = {
        "auth_type": kwargs.get("auth_type"),
        "cognito_profile": kwargs.get("cognito_profile"),
        "cognito_credentials_file": kwargs.get("cognito_credentials_file")
    }
    client_opts = {x: kwargs[x] for x in kwargs if x not in session_opts}
    return _get_default_session(**session_opts).client(*args, **client_opts)


def resource(*args, **kwargs):
    session_opts = {
        "auth_type": kwargs.get("auth_type"),
        "cognito_profile": kwargs.get("cognito_profile"),
        "cognito_credentials_file": kwargs.get("cognito_credentials_file")
    }
    resource_opts = {x: kwargs[x] for x in kwargs if x not in session_opts}
    return _get_default_session(**session_opts).resource(*args, **resource_opts)


# Patch boto3.session.Session so we can get our Cognito tokens from our session later
def parse_config(filename):
    config = load_config(filename)
    return config


@property
def id_token(self):
    return self.token_cache.tokens.get("id_token")


@property
def access_token(self):
    return self.token_cache.tokens.get("access_token")


@property
def refresh_token(self):
    return self.token_cache.tokens.get("refresh_token")


@property
def token_expires(self):
    return self.token_cache.tokens.get("token_expires")


@property
def cognito_tokens(self):
    return self.token_cache.tokens


botosession.id_token = id_token
botosession.access_token = access_token
botosession.refresh_token = refresh_token
botosession.token_expires = token_expires
botosession.tokens = cognito_tokens
botosession.token_cache = None



#!/usr/bin/env python3.8
from threading import Thread
from time import sleep
from io import StringIO
from os import environ
import datetime
import json
from logging import getLogger, INFO
from dateutil.tz import tzlocal
from dateutil.parser import parse
from boto3 import client, set_stream_logger
from botocore import UNSIGNED
from botocore.config import Config
from botocore.exceptions import ClientError
from botocore.credentials import CredentialProvider, RefreshableCredentials
from warrant.aws_srp import AWSSRP

set_stream_logger(name="botocore")
logger = getLogger("botocore")
logger.setLevel(INFO)


def get_cognito_config_from_env():
    envList = [
        "COGNITO_APP_ID",
        "COGNITO_PASSWORD",
        "COGNITO_USERNAME",
        "COGNITO_USER_POOL_ID",
    ]
    missing = [x for x in envList if x not in environ]
    for e in envList:
        if e not in environ:
            raise Exception(
                f"""
                It looks like you want to use Cognito credentials for role switching,
                but you are missing some environment variables. Missing:
                {e}
            """)
    if not missing:
        res = {
            "app_id": environ["COGNITO_APP_ID"],
            "password": environ["COGNITO_PASSWORD"],
            "username": environ["COGNITO_USERNAME"],
            "user_pool_id": environ["COGNITO_USER_POOL_ID"],
            "identity_pool_id": environ.get("COGNITO_IDENTITY_POOL_ID", " "),
            "role_arn": environ.get("COGNITO_ROLE_ARN") or environ.get("AWS_ROLE_ARN"),
            "auth_flow": environ.get("COGNITO_AUTH_FLOW", "enhanced"),
            "metadata": json.loads(environ.get("COGNITO_METADATA", "{}")),
            "role_expiry_time": int(environ.get("COGNITO_ROLE_EXPIRY_TIME", "900"))
        }
    else:
        res = {}
    return res


def get_cognito_config(config):
    opt_list = [
        "app_id",
        "password",
        "username",
        "user_pool_id"
    ]
    missing = [x for x in opt_list if x not in config]
    if missing and config:
        raise Exception(
            f"""
            It looks like you want to use Cognito credentials for role switching,
            but you are missing some variables from cognito_credentials file. Missing:
            {', '.join(missing)}
        """
        )
    return config


class TokenCache:
    def __init__(self, cache):
        self.cache = cache

    def cache_tokens(self, tokens):
        if isinstance(tokens.get("token_expires"), datetime.datetime):
            tokens["token_expires"] = str(tokens["token_expires"])

        if isinstance(self.cache, StringIO):
            self.cache.write(json.dumps(tokens))
            self.cache.seek(0)

        if isinstance(self.cache, str):
            with open(self.cache, "w") as f:
                json.dump(tokens, f)

    @property
    def tokens(self):
        try:
            if isinstance(self.cache, StringIO):
                self.cache.seek(0)
                tokens = json.loads(self.cache.getvalue())
            else:
                with open(self.cache, "r") as f:
                    tokens = json.load(f)
        except json.decoder.JSONDecodeError as e:
            if e.msg == "Expecting value":
                tokens = {}
            else:
                raise e

        return tokens


class TokenFetcher():
    def __init__(self, auth_type="user_srp", config={}, region_name=None, server=False, token_cache=None, non_blocking=False):
        if token_cache is None:
            io_obj = StringIO()
            token_cache = TokenCache(io_obj)
        else:
            token_cache = TokenCache(token_cache)

        self.provider = CognitoIdentity(auth_type=auth_type, config=config, region_name=region_name, token_cache=token_cache)

        self.provider.cognito_tokens = token_cache.tokens

        if non_blocking:
            Thread(target=self.provider.cognito_login, daemon=True)
        else:
            self.provider.cognito_login()

        if server:
            self.start_server()

    def fetch(self):
        self.provider.cognito_login()
        return self.provider.token_cache.tokens

    @property
    def tokens(self):
        return self.provider.token_cache.tokens

    @property
    def id_token(self):
        return self.provider.cognito_tokens["id_token"]

    @property
    def access_token(self):
        return self.provider.cognito_tokens["access_token"]

    @property
    def refresh_token(self):
        return self.provider.cognito_tokens["refresh_token"]

    @property
    def expires(self):
        return self.provider.cognito_tokens["token_expires"]

    def login_loop(self):
        while True:
            while datetime.datetime.now(tzlocal()) > parse(self.provider.cognito_tokens["token_expires"]) - datetime.timedelta(seconds=30):
                sleep(5)
            self.provider.cognito_login()

    def start_server(self):
        Thread(target=self.login_loop, daemon=True).start()


class CognitoIdentity(CredentialProvider):
    METHOD = 'cognito-identity'
    CANONICAL_NAME = 'customCognitoIdentity'
    api_credential_expiration = None
    cognito_tokens = {}
    auth = None
    tz = datetime.datetime.now(tzlocal())
    STS = None
    IDP = None

    def __init__(self, auth_type="user_srp", config={}, region_name=None, token_cache=None):
        super().__init__(self)
        self.token_cache = token_cache
        if config:
            self.config = get_cognito_config(config)
        else:
            self.config = get_cognito_config_from_env()
        self.config["region_name"] = region_name or config.get("region") or environ.get("AWS_DEFAULT_REGION")
        self.IDP = client(
            "cognito-idp",
            region_name=self.config["region_name"],
            config=Config(signature_version=UNSIGNED)
        )
        self.IDENTITY = client("cognito-identity", region_name=self.config["region_name"])
        if self.config["auth_flow"] == "classic":
            self.STS = client("sts")
            self.COGNITO_IDP = client("cognito-idp")

        self.cognito_tokens["refresh_token"] = None
        auth_type = auth_type or environ.get("COGNITO_AUTH_TYPE", "user_srp")
        if auth_type == "user_srp":
            self._auth_func = self._srp_auth
        elif auth_type == "user_password":
            self._auth_func = self._password_auth
        else:
            raise Exception("kwarg auth_type must be one of user_srp or user_password")

    def load(self):
        if self.config:
            fetcher = self._create_credentials_fetcher()
            credentials = fetcher(time_as="datetime")
            res = RefreshableCredentials(
                credentials["access_key"],
                credentials["secret_key"],
                credentials["token"],
                credentials["expiry_time"],
                refresh_using=fetcher,
                method=self.METHOD
            )
        else:
            res = None
        return res

    def _login(self):
        return self._auth_func()

    def refresh_auth(self):
        self.cognito_login()
        return self.cognito_tokens

    def cognito_login(self):
        logger.debug("Fetching Cognito credentials")
        try:
            if self.cognito_tokens.get("refresh_token"):
                auth = self._refresh_auth()
            else:
                auth = self._login()
            # Get the datetime that the token expires in - 1 minute just to be safe
            diff = auth["ExpiresIn"] - 1
            expires_in = datetime.datetime.now(tzlocal()) + datetime.timedelta(seconds=diff)

            self.cognito_tokens = {
                "id_token": auth["IdToken"],
                "access_token": auth["AccessToken"],
                "token_expires": str(expires_in),
                "refresh_token": auth["RefreshToken"]
            }

            self.token_cache.cache_tokens(self.cognito_tokens)
            return auth["IdToken"]

        except (Exception, ClientError) as e:
            logger.error(e)
            if self.cognito_tokens.get("refresh_token"):
                del self.cognito_tokens["refresh_token"]
            self.cognito_login()

    def _srp_auth(self):
        srp = AWSSRP(
            username=self.config["username"],
            password=self.config["password"],
            pool_id=self.config["user_pool_id"],
            client_id=self.config["app_id"],
            pool_region=self.config["region_name"],
        )

        AUTH_PARAMETERS = {
            "CHALLENGE_NAME": "SRP_A",
            "USERNAME": self.config["username"],
            "SRP_A": srp.get_auth_params()["SRP_A"],
        }

        auth = self.IDP.initiate_auth(
            AuthFlow="USER_SRP_AUTH",
            AuthParameters=AUTH_PARAMETERS,
            ClientId=self.config["app_id"],
            ClientMetadata=self.config.get("metadata"),
        )

        response = srp.process_challenge(auth["ChallengeParameters"])

        auth = self.IDP.respond_to_auth_challenge(
            ClientId=self.config["app_id"],
            ChallengeName="PASSWORD_VERIFIER",
            ChallengeResponses=response,
        )["AuthenticationResult"]

        return auth

    def _password_auth(self):
        AUTH_PARAMETERS = {
            "USERNAME": self.config["username"],
            "PASSWORD": self.config["password"]
        }
        auth = self.IDP.initiate_auth(
            AuthFlow="USER_PASSWORD_AUTH",
            AuthParameters=AUTH_PARAMETERS,
            ClientId=self.config["app_id"],
            ClientMetadata=self.config.get("metadata")
        )["AuthenticationResult"]

        self.auth = auth
        return auth

    def _refresh_auth(self):
        AUTH_PARAMETERS = {"REFRESH_TOKEN": self.token_cache.tokens["refresh_token"]}
        auth = self.IDP.initiate_auth(
            AuthFlow="REFRESH_TOKEN_AUTH",
            AuthParameters=AUTH_PARAMETERS,
            ClientId=self.config["app_id"],
            ClientMetadata=self.config.get("metadata"),
        )["AuthenticationResult"]

        auth["RefreshToken"] = self.cognito_tokens["refresh_token"]

        return auth

    def _create_credentials_fetcher(self):

        def fetch(time_as="string"):
            logger.debug("Checking credentials....")

            # Get a new idToken if this one has expired
            if self.token_cache.tokens.get("id_token") is None or datetime.datetime.now(tzlocal()) > parse(self.token_cache.tokens["token_expires"]):
                logger.debug("Retreiving new Cognito tokens.")
                id_token = self.cognito_login()
            else:
                id_token = self.token_cache.tokens["id_token"]

            try:
                identityId = self.IDENTITY.get_id(
                    IdentityPoolId=self.config["identity_pool_id"],
                    Logins={f"""cognito-idp.{self.config["region_name"]}.amazonaws.com/{self.config["user_pool_id"]}""": id_token}
                )["IdentityId"]

            except self.IDENTITY.exceptions.NotAuthorizedException as e:
                logger.info(e)
                self.token_cache.tokens["id_token"] = None
                fetch()

            if self.config["auth_flow"] == "classic":
                logger.debug("Using classic auth flow....")
                token = self.IDENTITY.get_open_id_token(
                    IdentityId=identityId,
                    Logins={
                        f"""cognito-idp.{self.config["region_name"]}.amazonaws.com/{self.config["user_pool_id"]}""": self.token_cache.tokens["id_token"]
                    }
                )["Token"]

                if not self.config.get("role_session_name"):
                    attributes = self.COGNITO_IDP.get_user(
                        AccessToken=self.token_cache.tokens["access_token"]
                    )["UserAttributes"]

                    self.config["role_session_name"] = [x["Value"] for x in attributes if x["Name"] == "sub"][0]

                opts = {
                    "WebIdentityToken": token,
                    "DurationSeconds": int(self.config.get("role_expiry_time", "900")),
                    "RoleSessionName": self.config["role_session_name"],
                    "RoleArn": self.config["role_arn"]
                }

                credentials = self.STS.assume_role_with_web_identity(**opts)
                credentials["Credentials"]["SecretKey"] = credentials["Credentials"]["SecretAccessKey"]
            else:
                opts = {
                    "IdentityId": identityId,
                    "Logins": {f"""cognito-idp.{self.config["region_name"]}.amazonaws.com/{self.config["user_pool_id"]}""": id_token}
                }
                if self.config.get("role_arn"):
                    opts["CustomRoleArn"] = self.config.get("role_arn")

                credentials = self.IDENTITY.get_credentials_for_identity(**opts)

            # We want to refresh whenever either the id token or iam is about to expire, whichever comes first
            if not self.token_cache.tokens.get("token_expires"):
                expire_time = credentials["Credentials"]["Expiration"]
            elif parse(self.token_cache.tokens["token_expires"]) < credentials["Credentials"]["Expiration"]:
                expire_time = parse(self.token_cache.tokens["token_expires"])
            else:
                expire_time = credentials["Credentials"]["Expiration"]

            # When we call load() expiry_time has to be a datetime, when we are called by RefreshableCredentials expiry_time needs
            # to be a string. I think its ugly myself. Open to suggestions.
            return {
                "access_key": credentials["Credentials"]["AccessKeyId"],
                "secret_key": credentials["Credentials"]["SecretKey"],
                "token": credentials["Credentials"]["SessionToken"],
                "expiry_time": str(expire_time) if time_as == "string" else expire_time
            }

        return fetch



import io
from setuptools import setup


setup(
    name="cognitoinator",
    version="0.0.21",
    description="The swiss army knife of cognito authentication",
    author="Mathew Moon",
    author_email="mmoon@quinovas.com",
    url="https://github.com/QuiNovas/cognitoinator",
    license="Apache 2.0",
    long_description=io.open("README.rst", encoding="utf-8").read(),
    long_description_content_type="text/x-rst",
    packages=["cognitoinator"],
    package_dir={"cognitoinator": "src/cognitoinator"},
    install_requires=["boto3", "botocore", "warrant"],
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
    ]
)
